# Largely placeholder file
# Created with some initial structure. Plans to refactor and simplify
# I prefer the use of initializers generally
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, Optional
from datetime import datetime
import time
import pandas as pd
import math

from .caching import load_df, load_schedule, load_drivers_df, save_drivers_df
from .schedule import Schedule
from .race import Race

def _nanmean(values):
    s = pd.to_numeric(values, errors="coerce").dropna()
    return float(s.mean()) if not s.empty else None

def _nansum(values):
    s = pd.to_numeric(values, errors="coerce").dropna()
    return float(s.sum()) if not s.empty else None

def _nanmedian(values):
    s = pd.to_numeric(values, errors="coerce").dropna()
    return float(s.median()) if not s.empty else None

def _drop_avg_columns(df: pd.DataFrame) -> pd.DataFrame:
    if not isinstance(df, pd.DataFrame) or df.empty:
        return df
    drop_cols = [c for c in df.columns if c.startswith("avg_")]
    return df.drop(columns=drop_cols, errors="ignore")

def _filter_by_race(df: pd.DataFrame, race_id: int) -> pd.DataFrame:
    """Safe filter by race_id; returns empty DataFrame if column missing."""
    if not isinstance(df, pd.DataFrame) or df.empty:
        return pd.DataFrame()
    if "race_id" not in df.columns:
        return pd.DataFrame()
    return df.loc[df["race_id"] == race_id]

@dataclass
class Driver:
    driver_id: int
    name: Optional[str] = None
    team: Optional[str] = None
    car_number: Optional[str] = None
    manufacturer: Optional[str] = None
    stats_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    adv_stats_df: pd.DataFrame = field(default_factory=pd.DataFrame)
    averages: pd.Series = field(default_factory=lambda: pd.Series(dtype="float64"))
    results_rows: pd.DataFrame = field(default_factory=pd.DataFrame)
    stage_rows: pd.DataFrame = field(default_factory=pd.DataFrame)
    lap_speed_summary: pd.DataFrame = field(default_factory=pd.DataFrame)
    pit_stops_df: pd.DataFrame = field(default_factory=pd.DataFrame)

    def _merge_cols_for_race(self, df: pd.DataFrame, race_id: int, cols: list, row: dict) -> bool:
        sub = _filter_by_race(df, race_id)
        if sub.empty:
            return False
        s0 = sub.iloc[0]
        added = False
        for c in cols:
            if c in s0.index and pd.notna(s0[c]):
                row[c] = s0[c]
                added = True
        return added

    def _pit_aggregates_for_race(self, race_id: int) -> pd.DataFrame:
        p = _filter_by_race(self.pit_stops_df, race_id)
        if p.empty:
            return pd.DataFrame()
        data = {
            "race_id": race_id,
            "total_pit_stops": int(len(p)),
            "avg_total_duration": _nanmean(p["total_duration"]) if "total_duration" in p.columns else None,
            "avg_pit_stop_duration": _nanmean(p["pit_stop_duration"]) if "pit_stop_duration" in p.columns else None,
            "avg_in_travel_duration": _nanmean(p["in_travel_duration"]) if "in_travel_duration" in p.columns else None,
            "avg_out_travel_duration": _nanmean(p["out_travel_duration"]) if "out_travel_duration" in p.columns else None,
            "avg_positions_gained_lost": _nanmean(p["positions_gained_lost"]) if "positions_gained_lost" in p.columns else None,
        }
        return pd.DataFrame([data])

    def get_race_row(self, race_id: int) -> Optional[dict]:
        row = {
            "race_id": race_id,
            "driver_id": self.driver_id,
            "driver_name": self.name,
            "team": self.team,
        }
        found_any = False

        # Determine participation
        r = _filter_by_race(self.results_rows, race_id)
        participated = False
        if not r.empty:
            r0 = r.iloc[0]
            laps_c = pd.to_numeric(r0.get("laps_completed"), errors="coerce")
            fin_pos = pd.to_numeric(r0.get("finishing_position"), errors="coerce")
            participated = (pd.notna(fin_pos)) or (pd.notna(laps_c) and laps_c > 0)
        if not participated and not _filter_by_race(self.lap_speed_summary, race_id).empty:
            participated = True
        if not participated and not _filter_by_race(self.pit_stops_df, race_id).empty:
            participated = True
        if not participated and not _filter_by_race(self.stage_rows, race_id).empty:
            participated = True
        if not participated:
            return None

        # Results merge
        found_any |= self._merge_cols_for_race(
            self.results_rows, race_id,
            [
                "qualifying_position", "qualifying_speed", "points", "playoff_points",
                "laps_completed", "starting_position", "finishing_position",
                "manufacturer", "car_number",
            ],
            row,
        )
        if "car_number" not in row and not r.empty and "driver_number" in r.columns:
            dn = r.iloc[0].get("driver_number")
            if pd.notna(dn):
                row["car_number"] = str(dn); found_any = True

        # Loop/advanced stats merge
        wanted_cols = [
            "mid_position", "closing_position", "closing_laps_diff",
            "best_position", "worst_position", "avg_position",
            "passes_green_flag", "passing_diff", "passed_green_flag",
            "quality_passes", "fast_laps", "top15_laps",
        ]
        found_any |= self._merge_cols_for_race(self.stats_df, race_id, wanted_cols, row)
        found_any |= self._merge_cols_for_race(self.adv_stats_df, race_id, wanted_cols, row)

        # No closing_position fallback

        # Backfill identifiers
        if "manufacturer" not in row and self.manufacturer is not None:
            row["manufacturer"] = self.manufacturer
        if "car_number" not in row and self.car_number is not None:
            row["car_number"] = self.car_number

        # Stage results
        row["stage1_position"] = 0; row["stage1_points"] = 0.0
        row["stage2_position"] = 0; row["stage2_points"] = 0.0
        s = _filter_by_race(self.stage_rows, race_id)
        if not s.empty and "stage" in s.columns:
            s1 = s.loc[s["stage"] == 1]
            if not s1.empty:
                r1 = s1.iloc[0]
                pos1 = pd.to_numeric(r1.get("position"), errors="coerce")
                pts1 = pd.to_numeric(r1.get("stage_points"), errors="coerce")
                row["stage1_position"] = int(pos1) if pd.notna(pos1) and int(pos1) <= 10 else 0
                row["stage1_points"] = float(pts1) if pd.notna(pts1) and row["stage1_position"] > 0 else 0.0
                found_any = True
            s2 = s.loc[s["stage"] == 2]
            if not s2.empty:
                r2 = s2.iloc[0]
                pos2 = pd.to_numeric(r2.get("position"), errors="coerce")
                pts2 = pd.to_numeric(r2.get("stage_points"), errors="coerce")
                row["stage2_position"] = int(pos2) if pd.notna(pos2) and int(pos2) <= 10 else 0
                row["stage2_points"] = float(pts2) if pd.notna(pts2) and row["stage2_position"] > 0 else 0.0
                found_any = True

        # Lap-speed summary merge
        found_any |= self._merge_cols_for_race(
            self.lap_speed_summary, race_id,
            ["leader_laps", "avg_speed_rank", "laps_completed_pct"],
            row,
        )

        # Pit aggregates merge
        pit_agg = self._pit_aggregates_for_race(race_id)
        found_any |= self._merge_cols_for_race(
            pit_agg, race_id,
            ["total_pit_stops", "avg_total_duration", "avg_pit_stop_duration",
             "avg_in_travel_duration", "avg_out_travel_duration", "avg_positions_gained_lost"],
            row,
        )

        # Ensure all requested keys are present
        ensure = wanted_cols + [
            "manufacturer", "car_number",
            "leader_laps", "avg_speed_rank", "laps_completed_pct",
            "total_pit_stops", "avg_total_duration", "avg_pit_stop_duration",
            "avg_in_travel_duration", "avg_out_travel_duration", "avg_positions_gained_lost",
            "qualifying_position", "qualifying_speed", "points", "playoff_points",
            "laps_completed", "starting_position", "finishing_position",
            "stage1_position", "stage1_points", "stage2_position", "stage2_points",
        ]
        for c in ensure:
            if c not in row:
                row[c] = None

        return row

    def add_stats(self, df: pd.DataFrame) -> None:
        if df is None or df.empty or "driver_id" not in df.columns:
            return
        rows = df[df["driver_id"] == self.driver_id]
        if rows.empty:
            return
        self.stats_df = pd.concat([self.stats_df, rows], ignore_index=True) if not self.stats_df.empty else rows.copy()
        if self.name is None and "driver_name" in rows.columns:
            n = rows["driver_name"].dropna()
            if not n.empty:
                self.name = str(n.iloc[0])

    def add_adv_stats(self, df: pd.DataFrame) -> None:
        if df is None or df.empty or "driver_id" not in df.columns:
            return
        rows = df[df["driver_id"] == self.driver_id]
        if rows.empty:
            return
        self.adv_stats_df = pd.concat([self.adv_stats_df, rows], ignore_index=True) if not self.adv_stats_df.empty else rows.copy()
        if self.name is None and "driver_name" in rows.columns:
            n = rows["driver_name"].dropna()
            if not n.empty:
                self.name = str(n.iloc[0])

    # add results-derived fields
    def add_results(self, res: pd.DataFrame, race_id: int) -> None:
        if not isinstance(res, pd.DataFrame) or res.empty or "driver_id" not in res.columns:
            return
        row = res[res["driver_id"] == self.driver_id]
        if row.empty:
            return

        # Skip non-starters (DNQ/DNS): no finishing_position and no laps completed
        try:
            r0 = row.iloc[0]
            laps_c = pd.to_numeric(r0.get("laps_completed"), errors="coerce")
            fin_pos = pd.to_numeric(r0.get("finishing_position"), errors="coerce")
            started = (pd.notna(fin_pos)) or (pd.notna(laps_c) and laps_c > 0)
            if not started:
                return
        except Exception:
            pass

        cols = [c for c in (
            "qualifying_position",
            "qualifying_speed",
            "points",
            "playoff_points",
            "laps_completed",
            "starting_position",
            "finishing_position",
            "driver",
            "team",
            "manufacturer",
            "driver_number",
        ) if c in row.columns]
        take = row[cols].copy()
        take = take.assign(race_id=race_id)
        if "driver_number" in take.columns and "car_number" not in take.columns:
            with pd.option_context('mode.chained_assignment', None):
                take["car_number"] = take["driver_number"].astype(str)
        self.results_rows = pd.concat([self.results_rows, take], ignore_index=True) if not self.results_rows.empty else take
        if self.name is None and "driver" in row.columns:
            n = row["driver"].dropna()
            if not n.empty:
                self.name = str(n.iloc[0])
        if self.team is None and "team" in row.columns:
            t = row["team"].dropna()
            if not t.empty:
                self.team = str(t.iloc[0])
        if self.car_number is None and "driver_number" in row.columns:
            dn = row["driver_number"].dropna()
            if not dn.empty:
                self.car_number = str(dn.iloc[0])
        if self.manufacturer is None and "manufacturer" in row.columns:
            m = row["manufacturer"].dropna()
            if not m.empty:
                self.manufacturer = str(m.iloc[0])

    # add stage 1/2 results (position, stage_points)
    def add_stage_results(self, stage1: Optional[pd.DataFrame], stage2: Optional[pd.DataFrame], race_id: int) -> None:
        frames = []
        if isinstance(stage1, pd.DataFrame) and not stage1.empty and "driver_id" in stage1.columns:
            r = stage1[stage1["driver_id"] == self.driver_id]
            if not r.empty:
                r = r[[c for c in ("position", "stage_points") if c in r.columns]].copy().assign(stage=1, race_id=race_id)
                frames.append(r)
        if isinstance(stage2, pd.DataFrame) and not stage2.empty and "driver_id" in stage2.columns:
            r = stage2[stage2["driver_id"] == self.driver_id]
            if not r.empty:
                r = r[[c for c in ("position", "stage_points") if c in r.columns]].copy().assign(stage=2, race_id=race_id)
                frames.append(r)
        if frames:
            add = pd.concat(frames, ignore_index=True)
            self.stage_rows = pd.concat([self.stage_rows, add], ignore_index=True) if not self.stage_rows.empty else add

    # compute leader_laps, avg_speed_rank, laps_completed_pct from laps cache
    def add_laps_for_race(self, laps: pd.DataFrame, res: pd.DataFrame, race_id: int) -> None:
        if not isinstance(laps, pd.DataFrame) or laps.empty:
            return
        if not isinstance(res, pd.DataFrame) or res.empty:
            return
        if "driver_number" not in res.columns or "driver_id" not in res.columns:
            return
        map_num_to_id = (
            res[["driver_number", "driver_id"]]
            .dropna()
            .assign(driver_number=res["driver_number"].astype(str))
            .drop_duplicates(subset=["driver_number"], keep="first")
            .set_index("driver_number")["driver_id"]
        )
        work = laps.copy()
        if "Number" not in work.columns or "lap_speed" not in work.columns or "Lap" not in work.columns:
            return
        work["driver_number"] = work["Number"].astype(str)
        work["lap_speed"] = pd.to_numeric(work["lap_speed"], errors="coerce")
        work["Lap"] = pd.to_numeric(work["Lap"], errors="coerce").astype("Int64")
        work["driver_id"] = work["driver_number"].map(map_num_to_id)
        work = work.dropna(subset=["driver_id", "Lap", "lap_speed"])
        if work.empty:
            return
        work["lap_speed_max"] = work.groupby("Lap")["lap_speed"].transform("max")
        work["lap_speed_rank"] = work.groupby("Lap")["lap_speed"].rank(ascending=False, method="min")
        mine = work[work["driver_id"] == self.driver_id]
        if mine.empty:
            return
        leader_laps = int((mine["lap_speed"] == mine["lap_speed_max"]).sum())
        avg_rank = float(mine["lap_speed_rank"].mean())
        total_race_laps = int(work["Lap"].max()) if work["Lap"].notna().any() else None
        laps_completed = int(mine["Lap"].count())
        laps_pct = (laps_completed / total_race_laps) if total_race_laps and total_race_laps > 0 else None
        summary = pd.DataFrame([{
            "race_id": race_id,
            "leader_laps": leader_laps,
            "avg_speed_rank": avg_rank,
            "laps_completed_pct": laps_pct
        }])
        self.lap_speed_summary = pd.concat([self.lap_speed_summary, summary], ignore_index=True) if not self.lap_speed_summary.empty else summary

    # collect pit stop rows and map to driver_id via results
    def add_pit_stops(self, pit: pd.DataFrame, res: pd.DataFrame, race_id: int) -> None:
        if not isinstance(pit, pd.DataFrame) or pit.empty:
            return
        if not isinstance(res, pd.DataFrame) or res.empty:
            return
        if "Driver" not in pit.columns or "driver" not in res.columns or "driver_id" not in res.columns:
            return

        # Build name map from results, stripping symbols for matching
        def _clean_name(name):
            if pd.isna(name):
                return None
            import re
            cleaned = str(name).strip()
            cleaned = re.sub(r'^[*#†‡§¶\s]+', '', cleaned)
            cleaned = re.sub(r'[*#†‡§¶\s]+$', '', cleaned)
            cleaned = re.sub(r'\s*\([^)]*\)\s*$', '', cleaned)
            return cleaned.strip()

        name_mappings = {
            "Daniel Suárez": "Daniel Suarez",
            "John H. Nemechek": "John Hunter Nemechek",
            "Ricky Stenhouse Jr": "Ricky Stenhouse Jr.",
        }

        def _normalize_name(name):
            cleaned = _clean_name(name)
            if cleaned in name_mappings:
                return name_mappings[cleaned]
            return cleaned

        # Create clean name maps from results (include number/manufacturer when present)
        res_cols = ["driver", "driver_id"]
        if "driver_number" in res.columns:
            res_cols.append("driver_number")
        if "manufacturer" in res.columns:
            res_cols.append("manufacturer")
        res_clean = res[res_cols].dropna(subset=["driver", "driver_id"]).copy()
        res_clean["clean_name"] = res_clean["driver"].apply(_normalize_name)

        name_to_id = res_clean.drop_duplicates("clean_name").set_index("clean_name")["driver_id"]
        name_to_num = res_clean.drop_duplicates("clean_name").set_index("clean_name")["driver_number"] if "driver_number" in res_clean.columns else None
        name_to_man = res_clean.drop_duplicates("clean_name").set_index("clean_name")["manufacturer"] if "manufacturer" in res_clean.columns else None

        # Map pit stops to driver_id + enrich with car_number/manufacturer
        work = pit.copy()
        work["clean_driver"] = work["Driver"].apply(_normalize_name)
        work["driver_id"] = work["clean_driver"].map(name_to_id)

        # car_number: prefer mapping from results
        if name_to_num is not None:
            work["car_number"] = work["clean_driver"].map(name_to_num).astype("string")
        # manufacturer: prefer pit's Manufacturer, else fallback to results
        if "Manufacturer" in work.columns:
            work["manufacturer"] = work["Manufacturer"]
        elif name_to_man is not None:
            work["manufacturer"] = work["clean_driver"].map(name_to_man)

        mine = work[work["driver_id"] == self.driver_id].copy()
        if mine.empty:
            return

        mine["race_id"] = race_id
        # Ensure consistent types
        if "car_number" in mine.columns:
            mine["car_number"] = mine["car_number"].astype("string")
        self.pit_stops_df = pd.concat([self.pit_stops_df, mine], ignore_index=True) if not self.pit_stops_df.empty else mine

    def total_races(self) -> int:
        race_count = len(self.results_rows)
        return race_count
        
    
    def compute_averages(self) -> None:
        parts = []
        if not self.stats_df.empty:
            parts.append(self.stats_df.select_dtypes(include=["number"]))
        if not self.adv_stats_df.empty:
            parts.append(self.adv_stats_df.select_dtypes(include=["number"]))
        if not parts:
            self.averages = pd.Series(dtype="float64")
            return

        merged = pd.concat(parts, axis=1)
        self.averages = merged.mean(numeric_only=True)

        # Stats-based helpers
        if not self.stats_df.empty:
            if "position" in self.stats_df.columns:
                self.averages["avg_finish_position"] = _nanmean(self.stats_df["position"])
            if "start_position" in self.stats_df.columns:
                self.averages["avg_start_position"] = _nanmean(self.stats_df["start_position"])
            if "rating" in self.stats_df.columns:
                self.averages["avg_rating"] = _nanmean(self.stats_df["rating"])
            if "fast_laps" in self.stats_df.columns:
                self.averages["avg_fast_laps"] = _nanmean(self.stats_df["fast_laps"])

        # Results-derived metrics
        if isinstance(self.results_rows, pd.DataFrame) and not self.results_rows.empty:
            for col, out_name in [
                ("qualifying_position", "avg_qualifying_position"),
                ("qualifying_speed", "avg_qualifying_speed"),
                ("points", "avg_points"),
                ("playoff_points", "avg_playoff_points"),
            ]:
                if col in self.results_rows.columns:
                    self.averages[out_name] = _nanmean(self.results_rows[col])

        # Stage metrics
        if isinstance(self.stage_rows, pd.DataFrame) and not self.stage_rows.empty:
            for stage in (1, 2):
                sub = self.stage_rows[self.stage_rows.get("stage") == stage]
                if not sub.empty:
                    if "stage_points" in sub.columns:
                        self.averages[f"avg_stage{stage}_points"] = _nanmean(sub["stage_points"])
                    if "position" in sub.columns:
                        self.averages[f"avg_stage{stage}_position"] = _nanmean(sub["position"])

        # Lap-speed summary metrics
        if isinstance(self.lap_speed_summary, pd.DataFrame) and not self.lap_speed_summary.empty:
            if "leader_laps" in self.lap_speed_summary.columns:
                self.averages["lap_speed_leader_laps_total"] = _nansum(self.lap_speed_summary["leader_laps"])
            if "avg_speed_rank" in self.lap_speed_summary.columns:
                self.averages["avg_lap_speed_rank"] = _nanmean(self.lap_speed_summary["avg_speed_rank"])
            if "laps_completed_pct" in self.lap_speed_summary.columns:
                self.averages["avg_laps_completed_pct"] = _nanmean(self.lap_speed_summary["laps_completed_pct"])

        # Pit stop totals and averages
        if isinstance(self.pit_stops_df, pd.DataFrame) and not self.pit_stops_df.empty:
            self.averages["total_pit_stops"] = int(len(self.pit_stops_df))
            for col, out_name in [
                # ("pit_in_race_time", "avg_pit_in_race_time"),
                # ("pit_out_race_time", "avg_pit_out_race_time"),
                ("total_duration", "avg_total_duration"),
                # ("box_stop_race_time", "avg_box_stop_race_time"),
                # ("box_leave_race_time", "avg_box_leave_race_time"),
                ("pit_stop_duration", "avg_pit_stop_duration"),
                ("in_travel_duration", "avg_in_travel_duration"),
                ("out_travel_duration", "avg_out_travel_duration"),
                ("positions_gained_lost", "avg_positions_gained_lost"),
            ]:
                if col in self.pit_stops_df.columns:
                    self.averages[out_name] = _nanmean(self.pit_stops_df[col])

        self.averages["total_races"] = self.total_races()


@dataclass
class DriversData:
    year: int
    series_id: int
    drivers: Dict[int, Driver] = field(default_factory=dict)
    summary_cache: Optional[pd.DataFrame] = None
    per_race_cache: Optional[pd.DataFrame] = None
    season_race_ids: list[int] = field(default_factory=list)

    @classmethod
    def build(
        cls,
        year: int,
        series_id: int,
        use_cache_only: bool = True,
        sleep_seconds: int = 10,
        use_drivers_cache: bool = True,
        reload_drivers_cache: bool = False,
    ):
        """
        Build DriversData for a season (cache-first):
        - Load schedule from cache; only fetch if allowed and missing.
        - Use only finished races from the schedule.
        - For each finished race, load cached driver_stats and driver_stats_advanced.
        - If both missing and fetching allowed, fetch, then re-load from cache (sleep between fetches).
        - Optionally use/save season-level drivers caches (summary and per-race).
        """
        out = cls(year=year, series_id=series_id)

        if use_drivers_cache and not reload_drivers_cache:
            summary = load_drivers_df(year=year, series_id=series_id, name="summary")
            per_race = load_drivers_df(year=year, series_id=series_id, name="per_race")
            if isinstance(summary, pd.DataFrame) and not summary.empty:
                out.summary_cache = summary
                out.per_race_cache = per_race if isinstance(per_race, pd.DataFrame) and not per_race.empty else None
                if isinstance(out.per_race_cache, pd.DataFrame) and "race_id" in out.per_race_cache.columns:
                    out.season_race_ids = (
                        pd.to_numeric(out.per_race_cache["race_id"], errors="coerce")
                        .dropna().astype(int).unique().tolist()
                    )
                else:
                    sched_df = load_schedule(year=year, series_id=series_id)
                    if isinstance(sched_df, pd.DataFrame) and not sched_df.empty and "race_id" in sched_df.columns:
                        now = pd.Timestamp.now(tz="UTC")
                        if "winner_driver_id" in sched_df.columns:
                            finished_df = sched_df[sched_df["winner_driver_id"].notna()]
                        elif "scheduled_at" in sched_df.columns:
                            finished_df = sched_df[pd.to_datetime(sched_df["scheduled_at"], errors="coerce", utc=True) <= now]
                        else:
                            finished_df = sched_df
                        out.season_race_ids = (
                            pd.to_numeric(finished_df["race_id"], errors="coerce")
                            .dropna().astype(int).unique().tolist()
                        )
                return out

        # Schedule: prefer cache
        sched_df = load_schedule(year=year, series_id=series_id)
        if (sched_df is None or sched_df.empty) and not use_cache_only:
            sched = Schedule(year, series_id)
            sched_df = sched.data

        if sched_df is None or sched_df.empty or "race_id" not in sched_df.columns:
            return out

        cached_sched = load_schedule(year=year, series_id=series_id)
        if (cached_sched is None or cached_sched.empty):
            if use_cache_only:
                return out
            sched = Schedule(year, series_id)            # may fetch, then cache
        else:
            sched = Schedule(year, series_id)            # loads from cache, no API hit

        finished_df = sched.get_finished_races()
        if finished_df is None or finished_df.empty or "race_id" not in finished_df.columns:
            return out

        race_ids = finished_df["race_id"].dropna().astype(int).tolist()
        out.season_race_ids = race_ids
        if not race_ids:
            return out
        
        if not race_ids:
            return out

        for race_id in race_ids:
            stats = load_df("driver_stats", year=year, series_id=series_id, race_id=race_id)
            adv = load_df("driver_stats_advanced", year=year, series_id=series_id, race_id=race_id)
            res = load_df("results", year=year, series_id=series_id, race_id=race_id)
            stage1 = load_df("stage_1_results", year=year, series_id=series_id, race_id=race_id)
            stage2 = load_df("stage_2_results", year=year, series_id=series_id, race_id=race_id)
            laps = load_df("laps", year=year, series_id=series_id, race_id=race_id)
            pit = load_df("pit_stops", year=year, series_id=series_id, race_id=race_id)

            if (stats is None or (isinstance(stats, pd.DataFrame) and stats.empty)) and \
               (adv is None or (isinstance(adv, pd.DataFrame) and adv.empty)):
                if use_cache_only:
                    continue
                Race(year, series_id, race_id, live=False)
                time.sleep(sleep_seconds)
                stats = load_df("driver_stats", year=year, series_id=series_id, race_id=race_id)
                adv = load_df("driver_stats_advanced", year=year, series_id=series_id, race_id=race_id)
                res = load_df("results", year=year, series_id=series_id, race_id=race_id)
                stage1 = load_df("stage_1_results", year=year, series_id=series_id, race_id=race_id)
                stage2 = load_df("stage_2_results", year=year, series_id=series_id, race_id=race_id)
                laps = load_df("laps", year=year, series_id=series_id, race_id=race_id)
                pit = load_df("pit_stops", year=year, series_id=series_id, race_id=race_id)
                missing_stats = (stats is None or (isinstance(stats, pd.DataFrame) and stats.empty))
                missing_adv = (adv is None or (isinstance(adv, pd.DataFrame) and adv.empty))
                if missing_stats and missing_adv:
                    continue

            if isinstance(stats, pd.DataFrame) and not stats.empty and "race_id" not in stats.columns:
                stats = stats.copy()
                stats["race_id"] = race_id
            if isinstance(adv, pd.DataFrame) and not adv.empty and "race_id" not in adv.columns:
                adv = adv.copy()
                adv["race_id"] = race_id

            driver_ids = set()
            if isinstance(stats, pd.DataFrame) and not stats.empty and "driver_id" in stats.columns:
                driver_ids.update(stats["driver_id"].dropna().astype(int).unique().tolist())
            if isinstance(adv, pd.DataFrame) and not adv.empty and "driver_id" in adv.columns:
                driver_ids.update(adv["driver_id"].dropna().astype(int).unique().tolist())

            for did in sorted(driver_ids):
                d = out.drivers.get(did)
                if d is None:
                    d = Driver(driver_id=did)
                    out.drivers[did] = d
                if isinstance(stats, pd.DataFrame) and not stats.empty:
                    d.add_stats(stats)
                if isinstance(adv, pd.DataFrame) and not adv.empty:
                    d.add_adv_stats(adv)
                if isinstance(res, pd.DataFrame) and not res.empty:
                    d.add_results(res, race_id)
                if isinstance(stage1, pd.DataFrame) or isinstance(stage2, pd.DataFrame):
                    d.add_stage_results(stage1, stage2, race_id)
                if isinstance(laps, pd.DataFrame) and not laps.empty and isinstance(res, pd.DataFrame) and not res.empty:
                    d.add_laps_for_race(laps, res, race_id)
                if isinstance(pit, pd.DataFrame) and not pit.empty and isinstance(res, pd.DataFrame) and not res.empty:
                    d.add_pit_stops(pit, res, race_id)

        for d in out.drivers.values():
            d.compute_averages()

        if use_drivers_cache:
            try:
                summary_df = out.to_dataframe()
                if isinstance(summary_df, pd.DataFrame) and not summary_df.empty:
                    save_drivers_df(summary_df, year=year, series_id=series_id, name="summary")
                per_race_rows = []
                for rid in race_ids:
                    for drv in out.drivers.values():
                        r = drv.get_race_row(rid)
                        if r:
                            per_race_rows.append(r)
                if per_race_rows:
                    per_race_df = pd.DataFrame(per_race_rows)
                    save_drivers_df(per_race_df, year=year, series_id=series_id, name="per_race")
                    # Keep in memory so the notebook can inspect it right away
                    out.per_race_cache = per_race_df.copy()
            except Exception:
                pass

        return out

    def to_dataframe(self,min_participation: float = 0.2) -> pd.DataFrame:
        """One row per driver with averages."""
        season_total = len(self.season_race_ids) if self.season_race_ids else None

        if isinstance(self.summary_cache, pd.DataFrame) and not self.summary_cache.empty and not self.drivers:
            results_df = self.summary_cache.copy()

            if min_participation > 0:
                threshold = max(1, math.ceil(min_participation * season_total))
                results_df = results_df[results_df["total_races"].fillna(0).astype(int) >= threshold]
            return results_df

        rows = []
        for did, d in self.drivers.items():
            row = {
                "driver_id": did,
                "driver_name": d.name,
                "team": d.team,                   # <— include team
                "car_number": d.car_number,
                "manufacturer": d.manufacturer,
            }
            if isinstance(d.averages, pd.Series) and not d.averages.empty:
                row.update({str(k): v for k, v in d.averages.items()})
            rows.append(row)

        results_df = pd.DataFrame(rows)
        
        if min_participation > 0:
            threshold = max(1, math.ceil(min_participation * season_total))
            results_df = results_df[results_df["total_races"].fillna(0).astype(int) >= threshold]
        return results_df
    
    def race_dataframe(self, race_id: int) -> pd.DataFrame:
        # Keep per-race averages; only drop season-level total_races
        if isinstance(self.per_race_cache, pd.DataFrame) and not self.per_race_cache.empty and not self.drivers:
            df = self.per_race_cache[self.per_race_cache.get("race_id") == race_id].copy()
            df = df.drop(columns=["total_races"], errors="ignore")
            return df.reset_index(drop=True)
        rows = []
        for d in self.drivers.values():
            r = d.get_race_row(race_id)
            if r:
                rows.append(r)
        df = pd.DataFrame(rows).drop(columns=["total_races"], errors="ignore")
        return df.reset_index(drop=True)

    def driver_races(self, driver_id: int) -> pd.DataFrame:
        """
        Return one row per race for this driver with per-race metrics (includes per-race averages).
        """
        d = self.drivers.get(driver_id)
        if d is not None:
            race_ids = []
            if isinstance(d.results_rows, pd.DataFrame) and "race_id" in d.results_rows.columns:
                race_ids = (
                    pd.to_numeric(d.results_rows["race_id"], errors="coerce")
                    .dropna().astype(int).unique().tolist()
                )
            rows = []
            for rid in sorted(race_ids):
                r = d.get_race_row(rid)
                if r:
                    rows.append(r)
            return pd.DataFrame(rows).sort_values("race_id", ignore_index=True)
        if isinstance(self.per_race_cache, pd.DataFrame) and not self.per_race_cache.empty:
            df = self.per_race_cache[self.per_race_cache["driver_id"] == driver_id].copy()
            return df.sort_values("race_id", ignore_index=True)
        return pd.DataFrame()

    # NEW: raw pit stops for a specific driver (optionally for a single race)
    def driver_pit_stops(self, driver_id: int, race_id: Optional[int] = None) -> pd.DataFrame:
        """
        Return raw pit stop rows for driver_id across races (or a single race if race_id provided).
        Includes all columns from the pit_stops cache plus driver_id/driver_name/race_id.
        """
        d = self.drivers.get(driver_id)
        # Prefer in-memory driver store
        if d is not None and isinstance(d.pit_stops_df, pd.DataFrame) and not d.pit_stops_df.empty:
            df = d.pit_stops_df.copy()
            if race_id is not None:
                df = df[df.get("race_id") == race_id]
            if "driver_id" not in df.columns:
                df["driver_id"] = driver_id
            if "driver_name" not in df.columns:
                df["driver_name"] = d.name
            # Ensure normalized manufacturer field exists
            if "manufacturer" not in df.columns and "Manufacturer" in df.columns:
                df["manufacturer"] = df["Manufacturer"]
            # Order helpful columns first
            first = [c for c in ["race_id", "driver_id", "driver_name", "car_number", "manufacturer"] if c in df.columns]
            rest = [c for c in df.columns if c not in first]
            return df[first + rest].sort_values(["race_id"], ignore_index=True)

        # Fallback: build from race caches (best-effort)
        rows = []
        for rid in (self.season_race_ids or []):
            if race_id is not None and rid != race_id:
                continue
            pit = load_df("pit_stops", year=self.year, series_id=self.series_id, race_id=rid)
            res = load_df("results", year=self.year, series_id=self.series_id, race_id=rid)
            if not isinstance(pit, pd.DataFrame) or pit.empty or not isinstance(res, pd.DataFrame) or res.empty:
                continue
            if "driver" not in res.columns or "driver_id" not in res.columns or "Driver" not in pit.columns:
                continue
            name_to_id = res[["driver", "driver_id"]].dropna().drop_duplicates("driver").set_index("driver")["driver_id"]
            name_to_num = res[["driver", "driver_number"]].dropna().drop_duplicates("driver").set_index("driver")["driver_number"] if "driver_number" in res.columns else None
            name_to_man = res[["driver", "manufacturer"]].dropna().drop_duplicates("driver").set_index("driver")["manufacturer"] if "manufacturer" in res.columns else None

            work = pit.copy()
            work["driver_id"] = work["Driver"].map(name_to_id)
            work = work[work["driver_id"] == driver_id].copy()
            if work.empty:
                continue
            work["race_id"] = rid
            work["driver_name"] = work["Driver"]
            if name_to_num is not None:
                work["car_number"] = work["Driver"].map(name_to_num).astype("string")
            if "Manufacturer" in work.columns:
                work["manufacturer"] = work["Manufacturer"]
            elif name_to_man is not None:
                work["manufacturer"] = work["Driver"].map(name_to_man)
            rows.append(work)
        if rows:
            df = pd.concat(rows, ignore_index=True)
            first = [c for c in ["race_id", "driver_id", "driver_name", "car_number", "manufacturer"] if c in df.columns]
            rest = [c for c in df.columns if c not in first]
            return df[first + rest].sort_values(["race_id"], ignore_index=True)
        return pd.DataFrame()

    # NEW: raw pit stops for a race across all drivers
    def race_pit_stops(self, race_id: int) -> pd.DataFrame:
        """
        Return raw pit stop rows for a race_id across all drivers.
        Includes all columns from pit_stops cache plus driver_id/driver_name.
        """
        pit = load_df("pit_stops", year=self.year, series_id=self.series_id, race_id=race_id)
        res = load_df("results", year=self.year, series_id=self.series_id, race_id=race_id)
        if not isinstance(pit, pd.DataFrame) or pit.empty:
            return pd.DataFrame()
        df = pit.copy()
        # Map names to ids/numbers/manufacturer when possible
        if isinstance(res, pd.DataFrame) and not res.empty and "driver" in res.columns and "driver_id" in res.columns and "Driver" in df.columns:
            name_to_id = res[["driver", "driver_id"]].dropna().drop_duplicates("driver").set_index("driver")["driver_id"]
            df["driver_id"] = df["Driver"].map(name_to_id)
            if "driver_number" in res.columns:
                name_to_num = res[["driver", "driver_number"]].dropna().drop_duplicates("driver").set_index("driver")["driver_number"]
                df["car_number"] = df["Driver"].map(name_to_num).astype("string")
            if "manufacturer" in res.columns and "manufacturer" not in df.columns:
                name_to_man = res[["driver", "manufacturer"]].dropna().drop_duplicates("driver").set_index("driver")["manufacturer"]
                df["manufacturer"] = df["Driver"].map(name_to_man)
        # Normalize manufacturer field to lower-case column name as alias
        if "manufacturer" not in df.columns and "Manufacturer" in df.columns:
            df["manufacturer"] = df["Manufacturer"]
        if "driver_name" not in df.columns and "Driver" in df.columns:
            df["driver_name"] = df["Driver"]
        df["race_id"] = race_id
        first = [c for c in ["race_id", "driver_id", "driver_name", "car_number", "manufacturer"] if c in df.columns]
        rest = [c for c in df.columns if c not in first]
        return df[first + rest].reset_index(drop=True)


# src/race.py

import pandas as pd
import requests
import warnings
from .codes import FLAG_CODE
from .caching import load_df, save_df


#https://cf.nascar.com/cacher/2023/2/5314/lap-times.json

class Race:
    def __init__(self, year, series_id,race_id=None,live=False):
        # Initialize race metadata
        self.year = year
        self.live = live
        self.series_id = series_id
        self.race_id = race_id
        self.entrants = None
        self.superspeedway = None
        self.winner = None
        self.data = {}

        # Practice and Qualifying.
        # This may be moved
        self.practice_data = pd.DataFrame()
        self.qualifying_data = pd.DataFrame()

        # Race Data
        self.name = None
        self.laps = None
        self.pit_stops = None
        self.events = None
        self.results = None
        self.cautions = None
        self.lead_changes = None
        self.lead_change_data = None
        self.start_time = None
        self.race_time = None
        self.end_time = None
        self.distance = None
        self.lap_count = None
        self.drivers = pd.DataFrame()
        self.driver_stats_advanced = pd.DataFrame()

        # Stage Results. Currently shows only the top 10
        self.stage_1_results = None
        self.stage_2_results = None
        self.stage_3_results = None
        self.s1_lap_count = None
        self.s2_lap_count = None
        self.s3_lap_count = None

        # Initialize the race data
        self.fetch_laps()
        self.get_pit_stops()
        self.fetch_events()
        self.get_race_data()
        self.get_driver_stats()
        self.get_advanced_driver_stats()


    def get_race_data(self):
        """
        Fetch data for the race
        """
        if not self.live:
            cached_results = load_df("results", year=self.year, series_id=self.series_id, race_id=self.race_id)
            cached_cautions = load_df("cautions", year=self.year, series_id=self.series_id, race_id=self.race_id)
            cached_lead_changes = load_df("lead_changes", year=self.year, series_id=self.series_id, race_id=self.race_id)
            cached_stage1 = load_df("stage_1_results", year=self.year, series_id=self.series_id, race_id=self.race_id)
            cached_stage2 = load_df("stage_2_results", year=self.year, series_id=self.series_id, race_id=self.race_id)
            cached_stage3 = load_df("stage_3_results", year=self.year, series_id=self.series_id, race_id=self.race_id)

            if cached_results is not None:
                self.results = cached_results
            if cached_cautions is not None:
                self.cautions = cached_cautions
            if cached_lead_changes is not None:
                self.lead_changes = len(cached_lead_changes) - 1 if not cached_lead_changes.empty else 0
                self.lead_change_data = cached_lead_changes

            if cached_stage1 is not None:
                self.stage_1_results = cached_stage1
            if cached_stage2 is not None:
                self.stage_2_results = cached_stage2
            if cached_stage3 is not None:
                self.stage_3_results = cached_stage3

            if (
                cached_results is not None
                and cached_cautions is not None
                and cached_lead_changes is not None
                and cached_stage1 is not None
                and cached_stage2 is not None
                and cached_stage3 is not None
            ):
                 print(f'Loading data from Cache for: {self.year}-{self.series_id}-{self.race_id}')
                 return
            
            url = f"https://cf.nascar.com/cacher/{self.year}/{self.series_id}/{self.race_id}/weekend-feed.json"
            print(f'Fetching data for Year:{self.year} Series: {self.series_id} Race: {self.race_id}')
        else:
            url = f"https://cf.nascar.com/cacher/live/series_{self.series_id}/{self.race_id}/weekend-feed.json"
            print(f'Loading live data for: Year:{self.year} Series: {self.series_id} Race: {self.race_id}')

        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            try:
                weekend = data.get('weekend_race', [])[0]
                weekend_runs = data.get('weekend_runs', [])
            except:
                weekend = {}
                weekend_runs = []

            self.race_time = weekend.get('total_race_time')
            self.distance = weekend.get('scheduled_distance')
            self.lap_count = weekend.get('scheduled_laps')
            self.s1_lap_count = weekend.get('stage_1_laps')
            self.s2_lap_count = weekend.get('stage_2_laps')
            self.s3_lap_count = weekend.get('stage_3_laps')
            self.entrants = weekend.get('number_of_cars_in_field')
            self.superspeedway = weekend.get('restrictor_plate')

            
            results = weekend.get('results', [])
            caution_segments = weekend.get('caution_segments', [])
            leaders = weekend.get('race_leaders', [])
            stages = weekend.get('stage_results', [])

            driver_res = []
            caution_rows = []
            leader_list = []
            stage_1 = []
            stage_2 = []
            stage_3 = []

            for i in results:
                driver_res.append({
                    'driver_id': i.get('driver_id'),
                    'driver': i.get('driver_fullname'),
                    'driver_number': i.get('car_number'),
                    'manufacturer': i.get('car_make'),
                    'sponsor': i.get('sponsor'),
                    'team': i.get('team_name'),
                    'team_id': i.get('team_id'),
                    'qualifying_order': i.get('qualifying_order'),
                    'qualifying_position': i.get('qualifying_position'),
                    'qualifying_speed': i.get('qualifying_speed'),
                    'starting_position': i.get('starting_position'),
                    'finishing_position': i.get('finishing_position'),                    
                    'laps_completed': i.get('laps_completed'),
                    'points': i.get('points_earned'),
                    'playoff_points': i.get('playoff_points_earned'),
                })
            self.results = pd.DataFrame(driver_res)
            if not self.live:
                save_df("results", self.results, year=self.year, series_id=self.series_id, race_id=self.race_id)

            for i in caution_segments:
                caution_rows.append({
                    'start_lap': i.get('start_lap'),
                    'end_lap': i.get('end_lap'),
                    'caution_type': i.get('reason'),
                    'comment': i.get('comment'),
                    'flag_state': i.get('flag_state'),
                })

            self.cautions = pd.DataFrame(caution_rows) if caution_rows else pd.DataFrame()
            self.cautions['Flag'] = self.cautions['flag_state'].map(FLAG_CODE) if not self.cautions.empty else None
            self.cautions['duration'] = self.cautions['end_lap'] - self.cautions['start_lap'] if not self.cautions.empty else None
            
            if not self.live:
                save_df("cautions", self.cautions, year=self.year, series_id=self.series_id, race_id=self.race_id)
            for i in leaders:
                leader_list.append({
                    'start_lap': i.get('start_lap'),
                    'end_lap': i.get('end_lap'),
                    'driver_name': None,
                    'car_number': i.get('car_number')
                })

            try:
                self.lead_changes = len(leader_list) - 1 if leader_list else 0
                self.lead_change_data = pd.DataFrame(leader_list) if leader_list else pd.DataFrame()
                self.lead_change_data['driver_name'] = self.lead_change_data['car_number'].map(self.results.set_index('driver_number')['driver']) if not self.lead_change_data.empty else None
                if not self.live:
                    save_df("lead_changes", self.lead_change_data, year=self.year, series_id=self.series_id, race_id=self.race_id)
            except Exception as e:
                print(f"Error processing lead changes: {e}")
            
            for i in stages:
                if i.get('stage_number') == 1:
                    for j in i.get('results', []):
                        stage_1.append({
                            'driver_id': j.get('driver_id'),
                            'driver_name': j.get('driver_fullname'),
                            'car_number': j.get('car_number'),
                            'position': j.get('finishing_position'),
                            'stage_points': j.get('stage_points'),
                        })
                elif i.get('stage_number') == 2:
                    for j in i.get('results', []):
                        stage_2.append({
                            'driver_id': j.get('driver_id'),
                            'driver_name': j.get('driver_fullname'),
                            'car_number': j.get('car_number'),
                            'position': j.get('finishing_position'),
                            'stage_points': j.get('stage_points'),
                        })
                elif i.get('stage_number') == 3:
                    for j in i.get('results', []):
                        stage_3.append({
                            'driver_id': j.get('driver_id'),
                            'driver_name': j.get('driver_fullname'),
                            'car_number': j.get('car_number'),
                            'position': j.get('finishing_position'),
                            'stage_points': j.get('stage_points'),
                        })

            self.stage_1_results = pd.DataFrame(stage_1) if stage_1 else pd.DataFrame()
            self.stage_2_results = pd.DataFrame(stage_2) if stage_2 else pd.DataFrame()
            self.stage_3_results = pd.DataFrame(stage_3) if stage_3 else pd.DataFrame()

            if not self.live:
                save_df('stage_1_results', self.stage_1_results, year=self.year, series_id=self.series_id, race_id=self.race_id)
                save_df('stage_2_results', self.stage_2_results, year=self.year, series_id=self.series_id, race_id=self.race_id)
                save_df('stage_3_results', self.stage_3_results, year=self.year, series_id=self.series_id, race_id=self.race_id)

            # Process weekend runs (practice and qualifying)
            for run in weekend_runs:
                name = (run.get('run_name') or '').lower()
                if 'practice' in name:
                    df = self.get_practice_results(run)
                    if 'practice 1' in name:
                        df['practice_number'] = 1
                    elif 'practice 2' in name:
                        df['practice_number'] = 2
                    elif 'practice 3' in name:
                        df['practice_number'] = 3
                    elif 'final practice' in name:
                        df['practice_number'] = 4
                    self.practice_data = pd.concat([self.practice_data, df], ignore_index=True) if not self.practice_data.empty else df
                elif 'pole qualifying' in name:
                    df = self.get_qualifying_results(run)
                    if 'round 1' in name:
                        df['qualifying_round'] = 1
                    elif 'final round' in name:
                        df['qualifying_round'] = 2
                    else:
                        df['qualifying_round'] = None
                    self.qualifying_data = pd.concat([self.qualifying_data, df], ignore_index=True) if not self.qualifying_data.empty else df
            if not self.live:
                if not self.practice_data.empty:
                    save_df("practice", self.practice_data, year=self.year, series_id=self.series_id, race_id=self.race_id)
                if not self.qualifying_data.empty:
                    save_df("qualifying", self.qualifying_data, year=self.year, series_id=self.series_id, race_id=self.race_id)

        else:
            print(f"Failed to retrieve race data: {response.status_code}")

    def fetch_laps(self):
        """Fetch lap times for the specified race ID."""
        if not self.live:
            cached = load_df("laps", year=self.year, series_id=self.series_id, race_id=self.race_id)
            if cached is not None:
                self.laps = cached
                return
        
        if self.live:
            url = f"https://cf.nascar.com/cacher/live/series_{self.series_id}/{self.race_id}/lap-times.json"
        else:
            url = f"https://cf.nascar.com/cacher/{self.year}/{self.series_id}/{self.race_id}/lap-times.json"

        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            self.data = data
            lap_times = []
            
            for i in data['laps']:
                driver = i.get('FullName')
                number = i.get('Number')
                manufacturer = i.get('Manufacturer')
                for j in i.get('Laps', []):
                    lap_times.append({
                        'Driver': driver,
                        'Number': number,
                        'Manufacturer': manufacturer,
                        'Lap': j.get('Lap'),
                        'lap_time': j.get('LapTime'),
                        'lap_speed': j.get('LapSpeed'),
                        'position': j.get('RunningPos'),
                    })
            self.laps = pd.DataFrame(lap_times)
            self.laps['Lap'] = pd.to_numeric(self.laps['Lap'], errors='coerce').astype('Int64')
            self.laps['lap_time'] = pd.to_timedelta(self.laps['lap_time'], errors='coerce')
            self.laps['lap_speed'] = pd.to_numeric(self.laps['lap_speed'], errors='coerce')

            # Save the DataFrame to cache
            if not self.live:
                save_df("laps", self.laps, year=self.year, series_id=self.series_id, race_id=self.race_id)
        else:
            print(f"Failed to retrieve lap data: {response.status_code}")
            print(f"url: {url}")

    def get_pit_stops(self):
        """Get pit stop information for the race."""
        if not self.live:
            cached_pit_stops = load_df("pit_stops", year=self.year, series_id=self.series_id, race_id=self.race_id)
            if cached_pit_stops is not None:
                self.pit_stops = cached_pit_stops
                return
            url = f"https://cf.nascar.com/cacher/{self.year}/{self.series_id}/{self.race_id}/live-pit-data.json"
        if self.live:
            url = f"https://cf.nascar.com/cacher/live/series_{self.series_id}/{self.race_id}/live-pit-data.json"
            
        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            stops = []
            for i in data:
                stops.append({
                    'Driver': i.get('driver_name'),
                    'Lap': i.get('lap_count'),
                    'Manufacturer': i.get('vehicle_manufacturer'),
                    'pit_in_flag_status': i.get('pit_in_flag_status'),
                    'pit_out_flag_status': i.get('pit_out_flag_status'),
                    'pit_in_race_time': i.get('pit_in_race_time'),
                    'pit_out_race_time': i.get('pit_out_race_time'),
                    'total_duration': i.get('total_duration'),
                    'box_stop_race_time': i.get('box_stop_race_time'),
                    'box_leave_race_time': i.get('box_leave_race_time'),
                    'pit_stop_duration': i.get('pit_stop_duration'),
                    'in_travel_duration': i.get('in_travel_duration'),
                    'out_travel_duration': i.get('out_travel_duration'),
                    'pit_stop_type': i.get('pit_stop_type'),
                    'left_front_tire_changed': i.get('left_front_tire_changed'),
                    'left_rear_tire_changed': i.get('left_rear_tire_changed'),
                    'right_front_tire_changed': i.get('right_front_tire_changed'),
                    'right_rear_tire_changed': i.get('right_rear_tire_changed'),
                    'previous_lap_time': i.get('previous_lap_time'),
                    'next_lap_time': i.get('next_lap_time'),
                    'pit_in_rank': i.get('pit_in_rank'),
                    'pit_out_rank': i.get('pit_out_rank'),
                    'positions_gained_lost': i.get('positions_gained_lost'),
                })
            # store both list and a DataFrame for convenience
            self.pit_stops = pd.DataFrame(stops) if stops else pd.DataFrame()
            if not self.live:
                save_df("pit_stops", self.pit_stops, year=self.year, series_id=self.series_id, race_id=self.race_id)
        else:
            print(f"Failed to retrieve pit stop data: {response.status_code}")
            print(f"url: {url}")

    def fetch_events(self):
        """ Fetch lap events and flags"""
        if not self.live:
            cached_events = load_df("events", year=self.year, series_id=self.series_id, race_id=self.race_id)
            if cached_events is not None:
                self.events = cached_events
                return
            url = f"https://cf.nascar.com/cacher/{self.year}/{self.series_id}/{self.race_id}/lap-notes.json"
        if self.live:
            url = f"https://cf.nascar.com/cacher/live/series_{self.series_id}/{self.race_id}/lap-notes.json"
            
        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            events = []
            laps = data.get('laps')
            
            for k,v in laps.items():
                for j in v:
                    events.append({
                        'Lap': k,
                        'Flag_State': j.get('FlagState'),
                        'Flag': None,
                        'note': j.get('Note'),
                        #'note_id': j.get('NoteID'),
                        'driver_ids': j.get('DriverIDs'),
                    })
            self.events = pd.DataFrame(events) if events else pd.DataFrame()
            if not self.events.empty:
                self.events['Flag'] = self.events['Flag_State'].map(FLAG_CODE)
            if not self.live:
                save_df("events", self.events, year=self.year, series_id=self.series_id, race_id=self.race_id)

        else:
            print(f"Failed to retrieve lap events: {response.status_code}")
            print(f"url: {url}")

    def get_practice_results(self, run: dict) -> pd.DataFrame:
        results = run.get('results', [])
        practice_res = []
        for i in results:
            practice_res.append({
                'driver_id': i.get('driver_id'),
                'driver_name': i.get('driver_name'),
                'manufacturer': i.get('manufacturer'),
                'position': i.get('finishing_position'),
                'lap_time': i.get('best_lap_time'),
                'speed': i.get('best_lap_speed'),
                'total_laps': i.get('laps_completed'),
                'delta_to_leader': i.get('delta_leader')
            })

        practice_data = pd.DataFrame(practice_res) if practice_res else pd.DataFrame()
        practice_data['practice_number'] = run.get('practice_number', 1)
        practice_data = practice_data.sort_values(by='position') if not practice_data.empty else pd.DataFrame()
        return practice_data

    def get_qualifying_results(self, run:dict) -> pd.DataFrame:
        results = run.get('results', [])
        quali_res = []
        for i in results:
            quali_res.append({
                'driver_id': i.get('driver_id'),
                'driver_name': i.get('driver_name'),
                'manufacturer': i.get('manufacturer'),
                'position': i.get('finishing_position'),
                'lap_time': i.get('best_lap_time'),
                'speed': i.get('best_lap_speed'),
                'total_laps': i.get('laps_completed'),
                'delta_to_leader': i.get('delta_leader')
            })
        qualifying_data = pd.DataFrame(quali_res) if quali_res else pd.DataFrame()
        qualifying_data['qualifying_round'] = None
        qualifying_data = qualifying_data.sort_values(by='position') if not qualifying_data.empty else pd.DataFrame()
        return qualifying_data

    def get_practice(self, round: int) -> pd.DataFrame:
        """Return practice data for a specific round."""
        data = self.practice_data[self.practice_data['practice_number'] == round] if not self.practice_data.empty else pd.DataFrame()
        return data

    def get_qualifying(self, round: int) -> pd.DataFrame:
        """Return qualifying data for a specific round."""
        data = self.qualifying_data[self.qualifying_data['qualifying_round'] == round] if not self.qualifying_data.empty else pd.DataFrame()
        return data
    
    def get_driver_stats(self):
        #https://cf.nascar.com/loopstats/prod/2023/2/5314.json
        #https://cf.nascar.com/cacher/live/series_2/5314/live-feed.json
        if not self.live:
            cached_driver_stats = load_df("driver_stats", year=self.year, series_id=self.series_id, race_id=self.race_id)
            if cached_driver_stats is not None:
                self.driver_stats = cached_driver_stats
                return
            
        url = f"https://cf.nascar.com/loopstats/prod/{self.year}/{self.series_id}/{self.race_id}.json"
        
        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()

            # Data comes in array
            drivers = data[0].get('drivers', [])
            driver_list = []
            for i in drivers:
                driver_list.append({
                    'driver_id': i.get('driver_id'),
                    'driver_name': None,
                    'start_position': i.get('start_ps'),
                    'mid_position': i.get('mid_ps'),
                    'position': i.get('ps'),
                    'closing_position': i.get('closing_ps'),
                    'closing_laps_diff': i.get('closing_laps_diff'),
                    'best_position': i.get('best_ps'),
                    'worst_position': i.get('worst_ps'),
                    'avg_position': i.get('avg_ps'),
                    'passes_green_flag': i.get('passes_gf'),
                    'passing_diff': i.get('passing_diff'),
                    'passed_green_flag': i.get('passed_gf'),
                    'quality_passes': i.get('quality_passes'),
                    'fast_laps': i.get('fast_laps'),
                    'top15_laps': i.get('top15_laps'),
                    'lead_laps': i.get('lead_laps'),
                    'laps': i.get('laps'),
                    'rating': i.get('rating'),
                })
            
            self.drivers = pd.DataFrame(driver_list) if driver_list else pd.DataFrame()
            if not self.drivers.empty and isinstance(self.results, pd.DataFrame) and not self.results.empty:
                name_map_df = (
                    self.results[['driver_id', 'driver']]
                    .dropna(subset=['driver_id', 'driver'])
                    .astype({'driver_id': 'Int64'})
                    .drop_duplicates(subset=['driver_id'], keep='first')
                )
                name_map = name_map_df.set_index('driver_id')['driver']
                self.drivers['driver_name'] = self.drivers['driver_id'].astype('Int64').map(name_map)
            if not self.live:
                save_df("driver_stats", self.drivers, year=self.year, series_id=self.series_id, race_id=self.race_id)

    def get_advanced_driver_stats(self):
        # Implement the logic to retrieve advanced driver stats
        if not self.live:
            cached_driver_stats_advanced = load_df("driver_stats_advanced", year=self.year, series_id=self.series_id, race_id=self.race_id)
            if cached_driver_stats_advanced is not None:
                self.driver_stats_advanced = cached_driver_stats_advanced
                return
            
        url = f"https://cf.nascar.com/cacher/live/series_{self.series_id}/{self.race_id}/live-feed.json"

        response = requests.get(url)
        if response.status_code == 200:
            data = response.json()
            # Process the advanced driver stats data
            vehicles = data.get('vehicles', [])
            driver_stats_advanced = []
            for vehicle in vehicles:
                driver = vehicle.get('driver', {})
                driver_stats_advanced.append({
                    'driver_id': driver.get('driver_id'),
                    'driver_name': driver.get('full_name'),
                    'vehicle_number': vehicle.get('vehicle_number'),
                    'vehicle_manufacturer': vehicle.get('vehicle_manufacturer'),
                    'sponsor_name': vehicle.get('sponsor_name'),
                    'best_lap': vehicle.get('best_lap'),
                    'best_lap_speed': vehicle.get('best_lap_speed'),
                    'best_lap_time': vehicle.get('best_lap_time'),
                    'laps_position_improved': vehicle.get('laps_position_improved'),
                    "fastest_laps_run": vehicle.get("fastest_laps_run"),
                    'passes_made': vehicle.get('passes_made'),
                    "times_passed": vehicle.get("times_passed"),
                    'passing_differential': vehicle.get('passing_differential'),
                    "quality_passes": vehicle.get("quality_passes"),
                    'position_differential_last_10_percent': vehicle.get('position_differential_last_10_percent'),
                })
            self.driver_stats_advanced = pd.DataFrame(driver_stats_advanced)
        else:
            self.driver_stats_advanced = pd.DataFrame()


from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import os

@dataclass
class Settings:
    # Only DataFrame caching now
    cache_enabled: bool = bool(os.getenv("PYNASCAR_CACHE_ENABLED", "1") not in ("0", "false", "False"))
    df_cache_enabled: bool = bool(os.getenv("PYNASCAR_DF_CACHE", "1") not in ("0", "false", "False"))
    cache_dir: Path = Path(os.getenv("PYNASCAR_CACHE_DIR", Path.home() / ".cache" / "pynascar")).expanduser()
    df_format: str = os.getenv("PYNASCAR_DF_FORMAT", "parquet")  # parquet|csv

_settings = Settings()

def get_settings() -> Settings:
    return _settings

def set_options(
        cache_enabled: bool | None = None,
        df_cache_enabled: bool | None = None,
        cache_dir: Path | str | None = None,
        df_format: str | None = None,
    ) -> Settings:
    """
    Configure DataFrame caching only. Supported formats: csv, parquet. No HTTP or SQL rn 
    """
    global _settings
    s = _settings

    if isinstance(cache_dir, str):
        cache_dir = Path(cache_dir)

    fmt = s.df_format if df_format is None else str(df_format).lower()
    if fmt not in ("csv", "parquet"):
        fmt = "parquet"
        raise UserWarning("Format must be csv or parquet. This will default to 'parquet'.")

    _settings = Settings(
        cache_enabled = s.cache_enabled if cache_enabled is None else cache_enabled,
        df_cache_enabled = s.df_cache_enabled if df_cache_enabled is None else df_cache_enabled,
        cache_dir = s.cache_dir if cache_dir is None else Path(cache_dir).expanduser(),
        df_format = fmt,
    )
    return _settings




tracks_map = {
    "Darlington Raceway": "intermediate",
    "World Wide Technology Raceway": "intermediate",
    "Homestead-Miami Speedway": "intermediate",
    "Bowman Gray Stadium": "short track",
    "Richmond Raceway": "short track",
    "Auto Club Speedway": "superspeedway",
    "Road America": "road course",
    "North Wilkesboro Speedway": "short track",
    "Las Vegas Motor Speedway": "intermediate",
    "Los Angeles Memorial Coliseum": "short track",
    "Sonoma Raceway": "road course",
    "Daytona International Speedway": "superspeedway",
    "Dover Motor Speedway": "intermediate",
    "Talladega Superspeedway": "superspeedway",
    "Rockingham Speedway": "intermediate",
    "Charlotte Motor Speedway": "intermediate",
    "Michigan International Speedway": "superspeedway",
    "Mid-Ohio Sports Car Course": "road course",
    "New Hampshire Motor Speedway": "short track",
    "DAYTONA Road Course": "road course",
    "Pocono Raceway": "superspeedway",
    "Knoxville Raceway": "short track",
    "Portland International Raceway": "road course",
    "Nashville Superspeedway": "intermediate",
    "Watkins Glen International": "road course",
    "Circuit of The Americas": "road course",
    "Iowa Speedway": "short track",
    "Dover International Speedway": "intermediate",
    "Lime Rock Park": "road course",
    "Bristol Motor Speedway": "short track",
    "Indianapolis Motor Speedway": "superspeedway",
    "Chicago Street Race": "road course",
    "Autódromo Hermanos Rodríguez": "road course",
    "Charlotte Motor Speedway Road Course": "road course",
    "Lucas Oil Indianapolis Raceway Park": "short track",
    "The Milwaukee Mile": "intermediate",
    "Indianapolis Motor Speedway Road Course": "road course",
    "Texas Motor Speedway": "intermediate",
    "Atlanta Motor Speedway": "intermediate",
    "Milwaukee Mile Speedway": "intermediate",
    "Kansas Speedway": "intermediate",
    "Bristol Motor Speedway Dirt": "short track",
    "Martinsville Speedway": "short track",
    "Phoenix Raceway": "short track"
}

tracks_by_type = {
    "superspeedway": [
        "Auto Club Speedway",
        "Daytona International Speedway",
        "Talladega Superspeedway",
        "Michigan International Speedway",
        "Pocono Raceway",
        "Indianapolis Motor Speedway"
    ],
    "intermediate": [
        "Darlington Raceway",
        "World Wide Technology Raceway",
        "Homestead-Miami Speedway",
        "Las Vegas Motor Speedway",
        "Rockingham Speedway",
        "Charlotte Motor Speedway",
        "Nashville Superspeedway",
        "Dover Motor Speedway",
        "Dover International Speedway",
        "The Milwaukee Mile",
        "Milwaukee Mile Speedway",
        "Texas Motor Speedway",
        "Atlanta Motor Speedway",
        "Kansas Speedway"
    ],
    "short track": [
        "Bowman Gray Stadium",
        "Richmond Raceway",
        "North Wilkesboro Speedway",
        "Los Angeles Memorial Coliseum",
        "New Hampshire Motor Speedway",
        "Knoxville Raceway",
        "Iowa Speedway",
        "Bristol Motor Speedway",
        "Lucas Oil Indianapolis Raceway Park",
        "Bristol Motor Speedway Dirt",
        "Martinsville Speedway",
        "Phoenix Raceway"
    ],
    "road course": [
        "Road America",
        "Sonoma Raceway",
        "Mid-Ohio Sports Car Course",
        "DAYTONA Road Course",
        "Portland International Raceway",
        "Watkins Glen International",
        "Circuit of The Americas",
        "Lime Rock Park",
        "Chicago Street Race",
        "Autódromo Hermanos Rodríguez",
        "Charlotte Motor Speedway Road Course",
        "Indianapolis Motor Speedway Road Course"
    ]
}


import re
import os
import warnings

# src/utils.py
# Utility functions for NASCAR data processing
# Can add regex patterns for validation or other utility functions as needed
def get_series_id(name):
    match name:
        case 'Cup Series':
            return 1
        case 'Xfinity':
            return 2
        case 'Truck Series':
            return 3
        case _:
            warnings.warn(f"Unknown series name: {name}, returning Cup Series ID")
            print(f'Options are: Cup Series, Xfinity, Truck Series')
            return 1
        
def get_series_name(series_id):
    match series_id:
        case 1:
            return 'Cup Series'
        case 2:
            return 'Xfinity'
        case 3:
            return 'Truck Series'
        case _:
            warnings.warn(f"Unknown series ID: {series_id}, returning Cup Series name")
            return 'Cup Series'





# src/pynascar/schedule.py
import warnings
import pandas as pd
import requests
from .caching import load_schedule, save_schedule
from .definitions import tracks_map

# endpoint for race list
#https://cf.nascar.com/cacher/2023/race_list_basic.json

class Schedule:
    '''
    
    Class to handle fetching and storing race data for a specific year and series ID.
    Attributes:
        year (int): The year of the race season.
        series_id (int): The ID of the NASCAR series (1 for Cup Series, 2 for Xfinity, 3 for Truck Series).
        
    '''
    
    def __init__(self, year, series_id):
        self.year = year
        self.series_id = series_id
        self.races = []
        self.data = pd.DataFrame()
        self.fetch_races()     

    def fetch_races(self):
        """Fetch the race list for the specified year and series ID."""
        cached_schedule = load_schedule(year=self.year, series_id=self.series_id)
        if cached_schedule is not None:
            self.data = cached_schedule
            self.races = self.data.to_dict(orient="records")
            return

        url = f"https://cf.nascar.com/cacher/{self.year}/race_list_basic.json"
        response = requests.get(url)
        if response.status_code == 200:
            print(f"Fetching data for Year:{self.year} Series:{self.series_id}")
            data = response.json()
            # Filter the races by series ID
            race_list = data[f'series_{self.series_id}']
            self.races = [race for race in race_list if race['series_id'] == self.series_id]
            self.data = pd.DataFrame(self.races)
            if "race_date" in self.data.columns:
                self.data["scheduled_at"] = pd.to_datetime(
                        self.data["race_date"], errors="coerce", utc=True
                    )
            
            self.data["track_type"] = self.data["track_name"].map(tracks_map).fillna("unknown")

            save_schedule(self.data, year=self.year, series_id=self.series_id)
        else:
            warnings.warn(f"Failed to fetch race list: {response.status_code}")
    
    def completed_races(self):
        """Return a list of completed races."""
        return self.get_finished_races()['race_name'].tolist(),self.get_finished_races()['race_id'].tolist()

    def remaining_races(self):
        """Return a list of remaining races."""
        return self.get_remaining_races()['race_name'].tolist(),self.get_remaining_races()['race_id'].tolist()

    def most_recent_race(self) -> pd.DataFrame:
        """Return a DataFrame with a single row for the most recent completed race."""
        return self.get_finished_races().head(1)
    
    def next_race(self) -> pd.DataFrame:
        """Return a DataFrame with a single row for the next scheduled race."""
        return self.get_remaining_races().head(1)

    def get_finished_races(self) -> pd.DataFrame:
        """
        Return a DataFrame with all completed races.
        """
        df = self.data.copy()
        if df.empty:
            return df

        now = pd.Timestamp.now(tz="UTC")
        completed = df["winner_driver_id"].notna() if "winner_driver_id" in df.columns else (df["scheduled_at"] <= now)
        df = df[completed]
        if df.empty:
            return df
        return df.sort_values("scheduled_at", ascending=False)
    
    def get_remaining_races(self) -> pd.DataFrame:
        """
        Return a DataFrame with all remaining races.
        """
        df = self.data.copy()
        if df.empty:
            return df

        now = pd.Timestamp.now(tz="UTC")
        remaining = df["winner_driver_id"].isna() if "winner_driver_id" in df.columns else (df["scheduled_at"] >= now)
        df = df[remaining]
        if df.empty:
            return df
        return df.sort_values("scheduled_at", ascending=True)

    


# This file is all boilerplate for functionality. NOT written by me. Can/will update later if needed 
# We are going with sqlite by default so we can use the same existing functions and just pass through the data
# This should work for now

from __future__ import annotations
from pathlib import Path
import re
import pandas as pd
from .config import get_settings

# Sanitize path segments and filenames
def _sanitize(s: str) -> str:
    return re.sub(r"[^A-Za-z0-9._-]+", "_", s).strip("_")

def _seg(v) -> str:
    if v is None:
        raise ValueError("year, series_id, and race_id must be provided")
    return _sanitize(str(v))

def race_cache_dir(year, series_id, race_id) -> Path:
    """
    <cache_dir>/<year>/<series_id>/<race_id>
    """
    s = get_settings()
    d = Path(s.cache_dir) / _seg(year) / _seg(series_id) / _seg(race_id)
    d.mkdir(parents=True, exist_ok=True)
    return d

def _cache_path(key: str, year, series_id, race_id, fmt: str | None = None) -> Path:
    s = get_settings()
    fmt = (fmt or s.df_format).lower()
    ext = ".parquet" if fmt == "parquet" else ".csv"
    return race_cache_dir(year, series_id, race_id) / f"{_sanitize(key)}{ext}"

def save_df(key: str, df: pd.DataFrame, *, year, series_id, race_id, fmt: str | None = None) -> Path:
    """
    Save: <cache_dir>/<year>/<series_id>/<race_id>/<key>.(csv|parquet)
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"save_df expects a pandas DataFrame; got {type(df).__name__}")
    s = get_settings()
    path = _cache_path(key, year, series_id, race_id, fmt)
    if not (s.cache_enabled and s.df_cache_enabled):
        return path  # no-op but return target path

    fmt = (fmt or s.df_format).lower()
    if fmt == "parquet":
        try:
            df.to_parquet(path, index=False)
        except Exception as e:
            raise RuntimeError(
                "Failed to write parquet. Install 'pyarrow' or set df_format='csv' via set_options()."
            ) from e
    elif fmt == "csv":
        df.to_csv(path, index=False)
    else:
        raise ValueError("Unsupported format. Use 'csv' or 'parquet'.")
    return path

def load_df(key: str, *, year, series_id, race_id, fmt: str | None = None) -> pd.DataFrame | None:
    """
    Load: <cache_dir>/<year>/<series_id>/<race_id>/<key>.(csv|parquet)
    """
    s = get_settings()
    if not (s.cache_enabled and s.df_cache_enabled):
        return None

    path = _cache_path(key, year, series_id, race_id, fmt)
    if not path.exists():
        return None

    fmt = (fmt or s.df_format).lower()
    if fmt == "parquet":
        try:
            return pd.read_parquet(path)
        except Exception as e:
            raise RuntimeError("Failed to read parquet. Install 'pyarrow' or use 'csv'.") from e
    elif fmt == "csv":
        return pd.read_csv(path)
    else:
        raise ValueError("Unsupported format. Use 'csv' or 'parquet'.")

def has_df(key: str, *, year, series_id, race_id, fmt: str | None = None) -> bool:
    return _cache_path(key, year, series_id, race_id, fmt).exists()

def clear_df(key: str, *, year, series_id, race_id, fmt: str | None = None) -> bool:
    p = _cache_path(key, year, series_id, race_id, fmt)
    if p.exists():
        p.unlink(missing_ok=True)
        return True
    return False

def clear_race(year, series_id, race_id) -> bool:
    """
    Delete <cache_dir>/<year>/<series_id>/<race_id> and prune empty parents.
    """
    d = race_cache_dir(year, series_id, race_id)
    removed = False
    if d.exists():
        for p in d.glob("*"):
            if p.is_file():
                p.unlink(missing_ok=True)
                removed = True
        try:
            d.rmdir()
        except OSError:
            pass
        # prune series and year if empty
        parent = d.parent
        for _ in range(2):
            try:
                parent.rmdir()
            except OSError:
                break
            parent = parent.parent
    return removed

def schedule_cache_dir(year) -> Path:
    """
    <cache_dir>/schedule/<year>
    """
    s = get_settings()
    d = Path(s.cache_dir) / "schedule" / _seg(year)
    d.mkdir(parents=True, exist_ok=True)
    return d

def _schedule_cache_path(year, series_id, fmt: str | None = None) -> Path:
    s = get_settings()
    fmt = (fmt or s.df_format).lower()
    ext = ".parquet" if fmt == "parquet" else ".csv"
    return schedule_cache_dir(year) / f"{_sanitize(str(series_id))}{ext}"

def save_schedule(df: pd.DataFrame, *, year, series_id, fmt: str | None = None) -> Path:
    """
    Save: <cache_dir>/schedule/<year>/<series_id>.(csv|parquet)
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"save_schedule expects a pandas DataFrame; got {type(df).__name__}")
    s = get_settings()
    path = _schedule_cache_path(year, series_id, fmt)
    if not (s.cache_enabled and s.df_cache_enabled):
        return path  # no-op

    fmt = (fmt or s.df_format).lower()
    if fmt == "parquet":
        try:
            df.to_parquet(path, index=False)
        except Exception as e:
            raise RuntimeError("Failed to write parquet. Install 'pyarrow' or set df_format='csv'.") from e
    elif fmt == "csv":
        df.to_csv(path, index=False)
    else:
        raise ValueError("Unsupported format. Use 'csv' or 'parquet'.")
    return path

def load_schedule(*, year, series_id, fmt: str | None = None) -> pd.DataFrame | None:
    """
    Load: <cache_dir>/schedule/<year>/<series_id>.(csv|parquet)
    """
    s = get_settings()
    if not (s.cache_enabled and s.df_cache_enabled):
        return None

    path = _schedule_cache_path(year, series_id, fmt)
    if not path.exists():
        return None

    fmt = (fmt or s.df_format).lower()
    if fmt == "parquet":
        try:
            return pd.read_parquet(path)
        except Exception as e:
            raise RuntimeError("Failed to read parquet. Install 'pyarrow' or use 'csv'.") from e
    elif fmt == "csv":
        return pd.read_csv(path)
    else:
        raise ValueError("Unsupported format. Use 'csv' or 'parquet'.")

def has_schedule(*, year, series_id, fmt: str | None = None) -> bool:
    return _schedule_cache_path(year, series_id, fmt).exists()

def clear_schedule(*, year, series_id, fmt: str | None = None) -> bool:
    p = _schedule_cache_path(year, series_id, fmt)
    if p.exists():
        p.unlink(missing_ok=True)
        return True
    return False

def drivers_cache_dir(year) -> Path:
    """
    <cache_dir>/drivers/<year>
    """
    s = get_settings()
    d = Path(s.cache_dir) / "drivers" / _seg(year)
    d.mkdir(parents=True, exist_ok=True)
    return d

def _drivers_cache_path(year, series_id, name: str, fmt: str | None = None) -> Path:
    s = get_settings()
    fmt = (fmt or s.df_format).lower()
    ext = ".parquet" if fmt == "parquet" else ".csv"
    return drivers_cache_dir(year) / f"{_sanitize(str(series_id))}_{_sanitize(name)}{ext}"

def save_drivers_df(df: pd.DataFrame, *, year, series_id, name: str, fmt: str | None = None) -> Path:
    """
    Save: <cache_dir>/drivers/<year>/<series_id>_<name>.(csv|parquet)
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"save_drivers_df expects a pandas DataFrame; got {type(df).__name__}")
    s = get_settings()
    path = _drivers_cache_path(year, series_id, name, fmt)
    if not (s.cache_enabled and s.df_cache_enabled):
        return path
    fmt = (fmt or s.df_format).lower()
    if fmt == "parquet":
        try:
            df.to_parquet(path, index=False)
        except Exception as e:
            raise RuntimeError("Failed to write parquet. Install 'pyarrow' or set df_format='csv'.") from e
    elif fmt == "csv":
        df.to_csv(path, index=False)
    else:
        raise ValueError("Unsupported format. Use 'csv' or 'parquet'.")
    return path

def load_drivers_df(*, year, series_id, name: str, fmt: str | None = None) -> pd.DataFrame | None:
    """
    Load: <cache_dir>/drivers/<year>/<series_id>_<name>.(csv|parquet)
    """
    s = get_settings()
    if not (s.cache_enabled and s.df_cache_enabled):
        return None
    path = _drivers_cache_path(year, series_id, name, fmt)
    if not path.exists():
        return None
    fmt = (fmt or s.df_format).lower()
    if fmt == "parquet":
        try:
            return pd.read_parquet(path)
        except Exception as e:
            raise RuntimeError("Failed to read parquet. Install 'pyarrow' or use 'csv'.") from e
    elif fmt == "csv":
        return pd.read_csv(path)
    else:
        raise ValueError("Unsupported format. Use 'csv' or 'parquet'.")

def has_drivers_df(*, year, series_id, name: str, fmt: str | None = None) -> bool:
    return _drivers_cache_path(year, series_id, name, fmt).exists()

def clear_drivers_df(*, year, series_id, name: str, fmt: str | None = None) -> bool:
    p = _drivers_cache_path(year, series_id, name, fmt)
    if p.exists():
        p.unlink(missing_ok=True)
        return True
    return False


# from .race import Race
# from .schedule import Schedule
# from .codes import FLAG_CODE
# from .utils import get_series_id, get_series_name
# from .config import get_settings, set_options

# __all__ = ["Race", "Schedule", "FLAG_CODE", "get_series_id", "get_series_name", "get_settings", "set_options"]



# 1-Green, 2-Yellow, 3-Red, 4-Finish, 6-Stop, 8-Warm Up, 9-Not Active
FLAG_CODE = {
    1: "Green",
    2: "Yellow",
    3: "Red",
    4: "Finish",
    6: "Stop",
    8: "Warm Up",
    9: "Not Active"
}


from .race import Race
from .schedule import Schedule
from .codes import FLAG_CODE
from .utils import get_series_id, get_series_name
from .config import get_settings, set_options

__all__ = ["Race", "Schedule", "FLAG_CODE", "get_series_id", "get_series_name", "get_settings", "set_options"]